ode_version::"0.02", ode_date::"17.08.1999";

/* Numerical solution of ode's (ordinary differential equations */

/* Equation: x'(t) = f(t,x), Initial value: x(t0) = x0; 

   num1ode       implements a naive algorithm (slightly improved Euler forward)
   rungekutta1   implements "classical" Runge Kutta algorithm 
*/

num1ode(f,t0,x0,step,fin):: *** Numerical solution for a first order ode ***
with                        *** step: stepsize (a small positive number)
  t::t0,                    *** fin: end of intervall to integrate to.
  x::x0,                    *** fin may be < t0, then step must be < 0.
  tx::0,                    *** Outputs a list of pairs (ti,x(ti)) in
  tplus::0,                 *** reverse order. The t values are always
  d::0,                     *** increased by step as long as t remains <= fin.
  l::list(list(t0,x0))      *** Due to rounding errors, this may occur one step
do                          *** earlier then expected, therefore use e.g.
                            *** num1ode(f,0,x0,0.001,1.0001) instead of
                            *** num1ode(f,0,x0,0.001,1). 
                            *** Use the reverse procedure to reverse the 
                            *** output order, if necesarry.
  if abs(step) < 1e-12 then 
    displayln("num1ode: step is zero or very close to zero."),
    list(first(l),first(l))
  elseif step*(fin-t0) < 0 then
    displayln("num1ode: t0 = ",t0," fin = ",fin," step = ",step,
                                                       " are incompatible."),
    list(first(l),first(l))
  elseif abs(t0-fin)<1e-12 then
    list(first(l),first(l))
  else
    while 
      tplus:=t+step,
      tx:=x+step*f(t,x),
      d:=(f(t,x)+f(tplus,tx))/2,
      x:=x+step*d,
      t:=tplus,
      if step>0 then
        tplus <= fin 
      else
        tplus >= fin
      endif
    do
      l := cons(list(t,x),l)
    endwhile,
    l
  endif
endwith;

rungekutta1(f,t0,x0,step,fin):: *** Numerical solution for a first order ode
with                            *** step: stepsize (a small positive number)
  t::t0,                        *** fin:  right end of intervall.
  x::x0,                        *** step must be < 0, if fin is < t0.
  tplus::0,                     *** Remarks on rounding errors apply as well.
  tplushalf::0,
  k1::0,
  k2::0,
  k3::0,
  k4::0,
  l::list(list(t0,x0))
do
  if abs(step) < 1e-12 then 
    displayln("num1ode: step is zero or very close to zero."),
    list(first(l),first(l))
  elseif step*(fin-t0) < 0 then
    displayln("num1ode: t0 = ",t0," fin = ",fin," step = ",step,
                                                       " are incompatible."),
    list(first(l),first(l))
  elseif abs(t0-fin)<1e-12 then
    list(first(l),first(l))
  else
    while 
      tplushalf:=t+step/2,
      tplus:=t+step,
      k1:=step*f(t,x),
      k2:=step*f(tplushalf,x+k1/2),
      k3:=step*f(tplushalf,x+k2/2),
      k4:=step*f(tplus,x+k3),
      x:=x+(k1+k2+k2+k3+k3+k4)/6,
      t:=tplus,
      if step>0 then
        tplus <= fin 
      else
        tplus >= fin
      endif
    do
      l := cons(list(t,x),l)
    endwhile,
    l
  endif
endwith;

/* System of two first order ode's  
   Equations:   x'(t) = f(t,x,y),
                y'(t) = g(t,x,y),  
Initial values: x(t0) = x0,
                y(t0) = y0,
                step    stepsize (step must be < 0, if fin < t0)
                fin     integrate from t0 to fin  

   num2ode       naive
   rungekutta2   Runge Kutta
*/

num2ode(f,g,t0,x0,y0,step,fin):: *** Numerical solution for a first order ode
with                             *** system of 2 equations
  t::t0,                         *** Remarks on rounding errors (see num1ode)
  x::x0,                         *** apply as well. Outputs a list of triples
  y::y0,                         *** (ti,x(ti),y(ti)) in reverse order.
  tx::0,
  ty::0,
  tplus::0,
  d::0,
  e::0,
  l::list(list(t0,x0,y0))
do
  if abs(step) < 1e-12 then 
    displayln("num2ode: step is zero or very close to zero."),
    list(first(l),first(l))
  elseif step*(fin-t0) < 0 then
    displayln("num2ode: t0 = ",t0," fin = ",fin," step = ",step,
                                                       " are incompatible."),
    list(first(l),first(l))
  elseif abs(t0-fin)<1e-12 then
    list(first(l),first(l))
  else
    while 
      tx:=x+step*f(t,x,y),
      ty:=y+step*g(t,x,y),
      tplus:=t+step,
      d:=(f(t,x,y)+f(tplus,tx,ty))/2,
      e:=(g(t,x,y)+g(tplus,tx,ty))/2,
      x:=x+step*d,
      y:=y+step*e,
      t:=tplus,
      if step>0 then 
        tplus <= fin
      else
        tplus >= fin
      endif
    do
      l := cons(list(t,x,y),l)
    endwhile,
    l
  endif
endwith;

rungekutta2(f,g,t0,x0,y0,step,fin):: *** Numerical solution for a first order
with                                 *** ode system of 2 equations
  t::t0,                             *** Remarks on rounding errors apply.
  x::x0,
  y::y0,
  ty::0,
  tplus::0,
  tplushalf::0,
  k1x::0,k1y::0,k2x::0,k2y::0,k3x::0,k3y::0,k4x::0,k4y::0,
  l::list(list(t0,x0,y0))
do
  if abs(step) < 1e-12 then 
    displayln("rungekutta2: step is zero or very close to zero."),
    list(first(l),first(l))
  elseif step*(fin-t0) < 0 then
    displayln("rungekutta2: t0 = ",t0," fin = ",fin," step = ",step,
                                                       " are incompatible."),
    list(first(l),first(l))
  elseif abs(t0-fin)<1e-12 then
    list(first(l),first(l))
  else
    while 
      tplus:=t+step,
      tplushalf:=t+step/2,
      k1x:=step*f(t,x,y),
      k1y:=step*g(t,x,y),
      k2x:=step*f(tplushalf, x+k1x/2, y+k1y/2),
      k2y:=step*g(tplushalf, x+k1x/2, y+k1y/2),
      k3x:=step*f(tplushalf, x+k2x/2, y+k2y/2),
      k3y:=step*g(tplushalf, x+k2x/2, y+k2y/2),
      k4x:=step*f(tplus,     x+k3x,   y+k3y),
      k4y:=step*g(tplus,     x+k3x,   y+k3y),
      x:=x+(k1x+k2x+k2x+k3x+k3x+k4x)/6,
      y:=y+(k1y+k2y+k2y+k3y+k3y+k4y)/6,
      t:=tplus,
      if step>0 then 
        tplus <= fin
      else
        tplus >= fin
      endif
    do
      l := cons(list(t,x,y),l)
    endwhile,
    l
  endif
endwith;

ode1o1n(f,t0,x0,step,t)::  *** Numerical solution for a first order ode, com-
                           *** putes solution for a special t > t0 or t < t0,
                           *** step is replaced by h so that h is close to
                           *** step and (t-x0)/h is an integer. x0 is the
                           *** initial value of the solution at t0. The
                           *** equation is  x'(t)=f(t,x). Internally,
                           *** rungekutta1 is used. Use the correct value
                           *** of t, the rounding error problem (see num1ode)
                           *** should not occur, since t is replaced by 
                           *** t+h/10 in the call of rungekutta1.
with h::0 do
  if abs(step)<1e-12 then
    displayln("ode1o1n: step is zero or very close to zero."),
    x0
  else
    h:abs(t-t0)/(1+truncate(abs((t-t0)/step))),
    if step < 0 then h: -h endif,
    second(first(rungekutta1(f,t0,x0,h,t+h/10)))
  endif
endwith;

ode2o1n(f,g,t0,x0,y0,step,t)::
                           *** Similiar to ode1o1n, however, a first order
                           *** system of two equations is solved:  
                           ***   Equations:   x'(t) = f(t,x,y),
                           ***                y'(t) = g(t,x,y),  
                           ***   Initial values: x(t0) = x0,
                           ***                   y(t0) = y0,
with h::0 do
  if abs(step)<1e-12 then
    displayln("ode2o1n: step is zero or very close to zero."),
    list(x0,y0)
  else
    h:abs(t-t0)/(1+truncate(abs((t-t0)/step))),
    if step < 0 then h:-h endif,
    rest(first(rungekutta2(f,g,t0,x0,y0,h,t+h/10)))
  endif
endwith;

ode1o2n(f,t0,x0,dx0,step,t)::
                           *** Solves the second order equation
                           ***  x''(t) = f(t,x,x') with initial values
                           ***  x(t0)  = x0, x'(t0) = dx0,
                           *** at t (t > t0 or t < t0).  
     first(ode2o1n((t,x,y)->y,f,t0,x0,dx0,step,t));

displayln("\nModule \"ode\" (Version ",ode_version,", ",ode_date,")");
